#include "parser.h"
#include "common.h" 
 
 
///////////////////////////////////////////////////////////////////////////////


const uint8_t CCParser::startUserDataHeader[]	= {0x00, 0x00, 0x01, 0xB2};
const uint8_t CCParser::endUserDataHeader[]	  = {0x00, 0x00, 0x01};
const uint8_t CCParser::dishDVBHeader[]		    = {0x05, 0x02}; 
const uint8_t CCParser::ATSCHeader[]		      = {0x47, 0x41, 0x39, 0x34}; 
const uint8_t CCParser::r4000DVBHeader[]	    = {0xBB, 0x02};
const uint8_t CCParser::r5000DVBHeader[]	    = {0x99, 0x02};
const uint8_t CCParser::dvdHeader[]		        = {0x43, 0x43, 0x01, 0xF8};


#define MAX_RETRIES 20


///////////////////////////////////////////////////////////////////////////////

 
CCParser::CCParser(CCManager* m, Interpreter* i) : cThread("CC Parser")
{
  dprint(L_DBG, "Parser  : Constructor.");
  manager = m;
  interpreter = i;
  
  inBuffer  = new cRingBufferLinear(KILOBYTE(256), 64, false, "CC Parser In");
  
  unknownUserData = 0;
  fPos = 0;
  dtvPos = 0;
  prevPictType = 0;
}


//-----------------------------------------------------------------------------

CCParser::~CCParser()
{
  Cancel(3);
  
  delete inBuffer; inBuffer = NULL;

  dprint(L_DBG, "Parser  : Destructor.");
}


//-----------------------------------------------------------------------------

void CCParser::Action(void) 
{
  dprint(L_DBG, "Parser  : Thread started.");
  int noStartFound=0;
  int noDelete=0;
  
  while (Running())
  {
    if (!inBuffer) break;
    
    int length=0;
    uint8_t* data = inBuffer->Get(length);
    if (data && length != 0)
    {
      //dprint(L_DBG, "Parser  : Got %d bytes from input buffer.", length);
      int used = 0;
      
      uint8_t* d = data;
      
      while (d - data < length-4)
      {
        uint8_t* start = NULL;
        while (d - data < length-4) {
          if (Equals(d, startUserDataHeader, 4)) {
            start = d;
            d+=4;
            break;
          }
          d++;  
        }
        if (!start) // Did not find start header
        {
          noStartFound++;
          //dprint(L_DBG, "Parser  : No start header found (%d)", noStartFound);
          used = length;
          break;
        }
      
        uint8_t* end = NULL;
        while (d - data < length-3) {
          if (Equals(d, endUserDataHeader, 3)) {
            end = d;
            break;
          }   
          d++;
        }
        
        if (!end)
        {
          used = start - data;
          dprint(L_DBG, "Parser  : No end found, used %d of %d", used, length);          
          break;
        }
        
        // Found both start and end
#if 0                         
          fprintf(stderr,"\n****************************************\n");
          for (uint8_t* dq = start; dq<end+4; dq++)
          fprintf(stderr,"%02X ", dq[0]);
          fprintf(stderr,"\n****************************************\n\n");
#endif 
        
        ProcessUserData(start+4, end-start-4);
        noStartFound = 0; 
      }
    
      dprint(L_DBG, "Parser  : Deleting %d of %d bytes from input buffer.", used, length);
      if (used == 0) {
        noDelete++;      
        dprint(L_DBG, "Parser  : Deleting 0 of %d bytes from input buffer. (%d)", length, noDelete);

        if (noDelete == MAX_RETRIES) {  
          dprint(L_ERR, "Parser  : %d bytes of stale data in buffer. Dumping.", length);
          used = length;
          noDelete=0;
        }    
      }
        
      inBuffer->Del(used);
      
    }
    else {
      noStartFound++;
      dprint(L_DBG, "Parser  : No data in input buffer (%d)", noStartFound);
    }
    
    if (noStartFound == MAX_RETRIES) {
      manager->NoDataFound();
    }  
       
    cCondWait::SleepMs(SLEEP_TIME);
  }
  
  
  dprint(L_DBG, "Parser  : Thread ended.");
}


//-----------------------------------------------------------------------------

void CCParser::ProcessUserData(uint8_t* data, int length)
{
  if (length < 2) return;
  //dprint(L_DBG, "Parser  : Processing %d bytes of user data.", length);
  if (Equals(data, dishDVBHeader, 2)) {
    ProcessDishCC(data+2, length-2);
    unknownUserData = 0; 
  }  
  else if (length>=4 && Equals(data, ATSCHeader,4)) {
    //dprint(L_MSG, "Parser  : CC Type ATSC Detected.");
    ProcessATSCCC(data+4, length-4);
    unknownUserData = 0; 
  }
  else if (Equals(data, r4000DVBHeader, 2)) {
    dprint(L_MSG, "Parser  : Unsuported CC Type Detected.");
    unknownUserData++;
  }
  else if (Equals(data, r5000DVBHeader, 2)) {
    dprint(L_MSG, "Parser  : Unsuported CC Type Detected.");
    unknownUserData++;
  }  
  else if (Equals(data, dvdHeader, 2)) {  
    dprint(L_MSG, "Parser  : Unsuported CC Type (DVD) Detected.");
    unknownUserData++;
  }
  else {
    dprint(L_MSG, "Parser  : Unknown user data header {0x%02X, 0x%02X}", data[0], data[1]);
    unknownUserData++;
  }
  
  if (unknownUserData == MAX_RETRIES) {
    manager->NoDataFound();
  }  
}


//-----------------------------------------------------------------------------

void CCParser::ProcessDishCC(uint8_t* data, int length)
{
  uint8_t* d = data+5;
  
  uint8_t type = d[0];
  
  if (type == 0x02 && d+2-data < length)
  {
    Store(d+2, 2);
    
    if( (d[4] == 0x04) && ((d[2] & 0x7F) < 0x20) )
      Store(d+2, 2);
  }
  
  else if (type == 0x04 && d+5-data < length)
  {
    Store(d+2, 4);
  }
  
  else if (type == 0x05 && d+10-data < length)
  {
    if (fPos>0) Store(fBuff, fPos);
    
    fPos = 0;

    fBuff[fPos++] = d[9];
    fBuff[fPos++] = d[10];
    
    if(d[7] == 0x02)
    {
      if( (d[11] == 0x04) && ((d[9] & 0x7F) < 0x20) ) {
        fBuff[fPos++] = d[9];
        fBuff[fPos++] = d[10];
      }
    }
    else
    {	
      fBuff[fPos++] = d[11];
      fBuff[fPos++] = d[12];				
		} 
		    
    d+=13;
  }
  else 
  { 
    dprint(L_ERR, "Parser  : Unknown data type 0x%02X at index %d of %d\n", type, d-data, length);  
  }
 
}


//-----------------------------------------------------------------------------

void CCParser::ProcessATSCCC(uint8_t* data, int length)
{
  uint8_t type = data[0] & 0x07; // The 0x07 because we added picture type to data[0]
  uint8_t pType = data[0] >> 3;
  
  if (type == 0x03)
  {
    //bool processEmData  = (data[1] >> 7) & 0x01;
    
    bool processCCData  = (data[1] >> 6) & 0x01;
    if (!processCCData) return;
    
    bool additionalData = (data[1] >> 5) & 0x01;
    uint8_t count = data[1] & 0x1F;
    //uint8_t emData = data[2];
     
    if (pType != 3 && prevPictType != pType) 
    { 
      Store(dtvBuffer, dtvPos);
      dtvPos=0; 
    }
    
    uint8_t* d = data + 3; 
    for (int i=0; i<count; i++)
    {
      bool ccValid = (d[0] >> 2) & 0x01; 
      uint8_t ccType = d[0] & 0x03;
      
      if (ccValid) {
        if ( (ccType == 0 || ccType == 1) && ccType == config.CCType)
        { 
          // EIA-608 captions          
          if (pType == 3) { // B frame
            Store(d+1, 2); 
          }  
          else // I or P frame
          { 
            if (d[1] != 0x80 || d[2] != 0x80) // Don't waste buffer space with
            {                                 // stuffing bytes
              dtvBuffer[dtvPos++] = d[1];
              dtvBuffer[dtvPos++] = d[2];
            
              //fprintf(stderr, "\n\ndtvPos: %d\n\n", dtvPos);
            }  
          }  
        }
        else if (ccType == 2 || ccType == 3) 
        {
          // EIA-708 captions
          dprint(L_MSG, "Parser  : Got EIA-708 captions. Not supported.");
        }
      }
      
      d+=3;
    }

   
    //data = d+1; // One byte marker

    if (additionalData) {
      // additional user data
      dprint(L_ERR, "Parser  : Additional user data ignored.");
    }            
  }
  else if (type == 0x04)
  {
    // Additional EIA 608 data
    dprint(L_ERR, "Parser  : Additional EIA 608 data ignored.");
  }
  else if (type == 0x05)
  {
    // luma PAM data
  }  
  else
  {
    // ATSC reserved user data
  }
  
  
  prevPictType = pType;
}


//-----------------------------------------------------------------------------

void CCParser::Store(uint8_t* data, int length)
{
  interpreter->Put(data, length);
}


//-----------------------------------------------------------------------------

void CCParser::Put(const uint8_t* data, int length)
{
  int p = inBuffer->Put(data, length);
  if (p != length) { 
    dprint(L_ERR, "Parser  : Input buffer overflow (%d bytes dropped)", length - p); 
    inBuffer->ReportOverflow(length - p);
  }
    
  // dprint(L_DBG, "Parser  : Stored %d of %d bytes", p, length);   
}


///////////////////////////////////////////////////////////////////////////////

